/*	Copyright (C) 2003-2015 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#include "operate/operate.pmh"

using namespace fe;
void OperateCommon::bind(sp<SignalerI> a_spSignalerI,
									sp<Layout> a_spLayout)
{
	m_aWindowEvent.initialize(a_spLayout->scope(),
			"WindowEvent");
}

//* static
void OperateCommon::drawLabel(sp<DrawI> a_spDrawI,
	const SpatialVector& a_location,String a_label,BWORD a_centered,
	U32 a_margin, const Color a_textColor,const Color* a_pBorderColor,
	const Color* a_pBackground)
{
	//* TODO query
	const U32 fontWidth=5;
	const U32 fontHeight=7;

	const U32 width=fontWidth*a_label.length();

	SpatialVector corner=a_location;
	if(a_centered)
	{
		corner-=0.5*SpatialVector(width,fontHeight);
		corner+=SpatialVector(0.5,0.5);
	}

	const SpatialVector behind=corner-SpatialVector(0.0,0.0,0.5);

	SpatialVector rect[2];
	rect[0]=behind-SpatialVector(a_margin-1,a_margin);
	rect[1]=behind+SpatialVector(width+a_margin,fontHeight+a_margin+1);

	Box2i backbox;
	set(backbox,rect[0][0]-2,rect[0][1]-2,
			rect[1][0]-rect[0][0]+4,rect[1][1]-rect[0][1]+4);

	if(a_pBackground)
	{
		a_spDrawI->drawRectangles(rect,2,false,a_pBackground);
	}

	if(a_pBorderColor)
	{
		a_spDrawI->drawBox(backbox,*a_pBorderColor);
	}

/*	text outline
	const Color black(0.0,0.0,0.0,1.0);
	const SpatialVector offsetX(1.0,0.0,0.0);
	const SpatialVector offsetY(0.0,1.0,0.0);
	a_spDrawI->drawAlignedText(corner+offsetX,a_label,black);
	a_spDrawI->drawAlignedText(corner-offsetX,a_label,black);
	a_spDrawI->drawAlignedText(corner+offsetY,a_label,black);
	a_spDrawI->drawAlignedText(corner-offsetY,a_label,black);
*/

	a_spDrawI->drawAlignedText(corner,a_label,a_textColor);
}

//* static
void OperateCommon::drawPicker(sp<DrawI> a_spDrawI,
	const SpatialVector& a_point,const SpatialVector& a_up,
	const SpatialVector& a_normal,Real a_scale,I32 a_depth,
	const Color* a_pColor)
{
	SpatialTransform transform;
	makeFrameNormalY(transform,a_point,a_up,a_normal);
	rotate(transform,-90.0*degToRad,e_xAxis);

	SpatialVector scale(a_scale,a_scale,a_scale);

//	a_spDrawI->drawCircle(transform,&scale,*a_pColor);
	//* HACK arc forces wire
	a_spDrawI->drawArc(transform,&scale,0.0,2.0*fe::pi,*a_pColor);

	for(I32 m=1;m<=a_depth;m++)
	{
		SpatialVector line[4];
		line[0]=SpatialVector(-1.0-0.5*m,-0.5);
		line[1]=SpatialVector(-1.0-0.5*m,0.5);
		line[2]=SpatialVector(1.0+0.5*m,-0.5);
		line[3]=SpatialVector(1.0+0.5*m,0.5);
		a_spDrawI->drawTransformedLines(transform,&scale,line,NULL,4,
				DrawI::e_discrete,false,a_pColor);
	}
}

//* static
void OperateCommon::drawImpact(sp<DrawI> a_spDrawI,
	sp<SurfaceTriangles::Impact> a_spTriImpact,
	Real a_inset,const Color* a_pColor)
{
	drawInsetTriangle(a_spDrawI,
			a_spTriImpact->vertex0(),
			a_spTriImpact->vertex1(),
			a_spTriImpact->vertex2(),
			a_inset,a_pColor);
}

//* static
void OperateCommon::drawInsetTriangle(sp<DrawI> a_spDrawI,
	const SpatialVector& a_vertex0,
	const SpatialVector& a_vertex1,
	const SpatialVector& a_vertex2,
	Real a_inset,const Color* a_pColor)
{
	const Real scale=1.0-2.0*a_inset;

	SpatialVector line[4];
	line[0]=scale*a_vertex0+a_inset*a_vertex1+a_inset*a_vertex2;
	line[1]=a_inset*a_vertex0+scale*a_vertex1+a_inset*a_vertex2;
	line[2]=a_inset*a_vertex0+a_inset*a_vertex1+scale*a_vertex2;
	line[3]=line[0];

	a_spDrawI->drawLines(line,NULL,4,DrawI::e_strip,false,a_pColor);
}

//* static
void OperateCommon::highlightCurve(sp<ViewI> a_spView,
	sp<DrawI> a_spDraw,sp<SurfaceAccessorI> a_spVertices,
	I32 a_primitiveIndex,I32 a_vertexIndex,
	const Color* a_pLineColor,const Color* a_pPointColor,
	const Color* a_pPointHighlight)
{
	const I32 subCount=a_spVertices->subCount(a_primitiveIndex);
	SpatialVector line[subCount];

	for(I32 subIndex=0;subIndex<subCount;subIndex++)
	{
		const SpatialVector point=
				a_spVertices->spatialVector(a_primitiveIndex,subIndex);

		const Vector2i projected=
				a_spView->project(point,ViewMode::e_perspective);

		line[subIndex]=projected;	//* NOTE z defaults to 0.0
	}

	if(a_pLineColor)
	{
		a_spDraw->drawLines(line,NULL,subCount,
				DrawI::e_strip,FALSE,a_pLineColor);
	}

	if(a_pPointColor)
	{
		for(I32 subIndex=0;subIndex<subCount;subIndex++)
		{
			line[subIndex][2]=0.5;
		}

		a_spDraw->drawPoints(line,NULL,subCount,FALSE,a_pPointColor);
	}

	if(a_vertexIndex>=0 && a_pPointHighlight)
	{
		const SpatialVector point=
				a_spVertices->spatialVector(a_primitiveIndex,a_vertexIndex);
		drawDot(a_spView,a_spDraw,point,5.0,*a_pPointHighlight);
	}
}

//* static
void OperateCommon::drawDot(sp<ViewI> a_spView,
	sp<DrawI> a_spDraw,SpatialVector a_point,Real a_radius,const Color a_color)
{
		const Vector2i projected=
				a_spView->project(a_point,ViewMode::e_perspective);
		SpatialVector center(projected[0],projected[1],0.6);

		SpatialTransform transform;
		setIdentity(transform);
		translate(transform,center);

		const SpatialVector scale(a_radius,a_radius,a_radius);
		a_spDraw->drawCircle(transform,&scale,a_color);
}

//* static
void OperateCommon::drawSquare(sp<DrawI>& a_rspDraw,
	const SpatialVector& a_rPoint,Real a_rimSize,
	const Color& a_rColorCenter,const Color& a_rColorRim,
	const Color* a_pColorShadow,const Color* a_pColorSparkle)
{
	sp<ViewI> spView=a_rspDraw->view();

	const Vector2i projected=spView->project(a_rPoint,ViewMode::e_perspective);
	const Vector2i rim(a_rimSize,a_rimSize);

	Box2i box=projected-rim;
	box.size()=2.0*rim+Vector2i(1,1);

	if(a_rColorCenter[3]>0.0)
	{
		const I32 lineWidth=a_rspDraw->drawMode()->lineWidth();

		SpatialVector corner[2];
		corner[0]=SpatialVector(box)+SpatialVector(lineWidth,lineWidth);
		corner[1]=SpatialVector(box)-SpatialVector(lineWidth,lineWidth)+
				SpatialVector(box.size());

		a_rspDraw->drawRectangles(corner,2,FALSE,&a_rColorCenter);
	}

	if(a_rColorRim[3]>0.0)
	{
		a_rspDraw->drawBox(box,a_rColorRim);
	}

	if(a_pColorShadow && (*a_pColorShadow)[3]>0.0)
	{
		SpatialVector glint=SpatialVector(box)+
				SpatialVector(box.size()[0],0.0,1.0);
		a_rspDraw->drawPoints(&glint,NULL,1,FALSE,a_pColorShadow);
	}

	if(a_pColorSparkle && (*a_pColorSparkle)[3]>0.0)
	{
		SpatialVector glint=SpatialVector(box)+
				SpatialVector(0.0,box.size()[1],1.0);
		a_rspDraw->drawPoints(&glint,NULL,1,FALSE,a_pColorSparkle);
	}
}

//* static
void OperateCommon::drawTube(sp<DrawI>& a_rspDraw,
	const SpatialVector* a_point,
	const SpatialVector* a_normal,
	const Real* a_radius,
	U32 a_pointCount,CurveMode a_curveMode,
	BWORD a_multicolor,const Color *a_color,
	const SpatialVector& a_rTowardCamera)
{
	const Color black(0.0,0.0,0.0,1.0);
	const Color white(1.0,1.0,1.0,1.0);

	const U32 slices=8;	//* TODO param

	const SpatialVector offset=1e-2*a_rTowardCamera;

	if(a_curveMode==e_lines)
	{
		//* overlapping wide lines

		SpatialVector line[a_pointCount];
		Color color[a_pointCount];
		Color glint[a_pointCount];

		for(U32 pointIndex=0;pointIndex<a_pointCount;pointIndex++)
		{
			line[pointIndex]=a_point[pointIndex]+offset;

			if(a_color)
			{
				color[pointIndex]=a_color[a_multicolor? pointIndex: 0];
			}
			else
			{
				color[pointIndex]=white;
			}

			glint[pointIndex]=color[pointIndex];

			//* TODO restore
//			glint[pointIndex]+=
//					(1.0-darken)*(white-color[pointIndex]);
			glint[pointIndex]=white;

			glint[pointIndex][3]=1.0;
		}

		for(U32 pass=0;pass<3;pass++)
		{
			//* TODO restore line changing
//			a_rspDraw->pushDrawMode(m_spWideLine[pass*2]);

			a_rspDraw->drawLines(line,NULL,a_pointCount,
					DrawI::e_strip,(pass!=2),
					pass? (pass==2? &black: color): glint);

//			a_rspDraw->popDrawMode();
		}
		return;
	}

	//* camera-facing tri-strip
	//* continuous tubes
	//* disjoint cylinders

	const U32 vertexCount=(a_curveMode==e_panels)? a_pointCount*2: slices*2+2;
	SpatialVector vertex[vertexCount];
	SpatialVector vertex2[vertexCount];
	SpatialVector vertexNormal[vertexCount];
	SpatialVector vertexNormal2[vertexCount];
	Color vertexColor[vertexCount];

	SpatialVector lastPoint=a_point[(a_curveMode==e_panels)? 1: 0];
	for(U32 pointIndex=(a_curveMode==e_panels? 0: 1);
			pointIndex<a_pointCount;pointIndex++)
	{
		const SpatialVector point=a_point[pointIndex];

		const Color color=a_color? a_color[a_multicolor? pointIndex: 0]: white;
		const Real radius=a_radius? a_radius[pointIndex]: 1.0;

		if(a_curveMode==e_cylinders)
		{
			a_rspDraw->drawCylinder(lastPoint,point,radius,radius,color,slices);
		}
		else
		{
			SpatialVector normalY(0.0,1.0,0.0);
			if(a_curveMode==e_panels)
			{
				//* TODO maybe faster in 2D screen space

				normalY=a_rTowardCamera;
			}
			else if(a_normal)
			{
				normalY=a_normal[pointIndex];
			}

			const SpatialVector delta=(point-lastPoint)*(pointIndex? 1: -1);
			const Real segLength=magnitude(delta);
			const SpatialVector tangentX=delta*
					(segLength>0.0? 1.0/segLength: 1.0);
			const SpatialVector offsetPoint=point+offset;

			SpatialTransform segmentFrame;
			makeFrameTangentX(segmentFrame,offsetPoint,tangentX,normalY);
			const SpatialVector sideZ=segmentFrame.column(2);

			if(a_curveMode==e_panels)
			{
				//* TODO use original uncorrected normals if provided

				const SpatialVector renormal=a_normal?
						a_normal[pointIndex]: segmentFrame.column(1);

				U32 vertexIndex=pointIndex*2;
				vertexColor[vertexIndex]=color;

//								transformVector(segmentFrame,
//										-v0,vertex[vertexIndex]);
				vertex[vertexIndex]=offsetPoint-radius*sideZ;

//								rotateVector(segmentFrame,
//										n0,vertexNormal[vertexIndex]);
				vertexNormal[vertexIndex]=sideZ;

				vertex2[vertexIndex]=offsetPoint;

//								rotateVector(segmentFrame,
//										n1,vertexNormal2[vertexIndex]);
				vertexNormal2[vertexIndex]=renormal;

				vertexIndex++;
				vertexColor[vertexIndex]=color;

				vertex[vertexIndex]=offsetPoint;

				vertexNormal[vertexIndex]=renormal;

//								transformVector(segmentFrame,
//										v0,vertex2[vertexIndex]);
				vertex2[vertexIndex]=offsetPoint+radius*sideZ;

//								vertexNormal2[vertexIndex]=
//										-vertexNormal[vertexIndex-1];
				vertexNormal2[vertexIndex]= -sideZ;
/*
				feLog("panel %d/%d %d/%d  %s  %s n %s  %s\n",
						primitiveIndex,primitiveCount,
						pointIndex,a_pointCount,
						c_print(vertex[vertexIndex-1]),
						c_print(vertex[vertexIndex]),
						c_print(vertexNormal[vertexIndex-1]),
						c_print(vertexNormal[vertexIndex]));
*/
			}
			else
			{
				const BWORD copying=(pointIndex>1);

				if(copying)
				{
					for(U32 vertexIndex=0;vertexIndex<vertexCount;
							vertexIndex+=2)
					{
						vertex[vertexIndex]=vertex[vertexIndex+1];
						vertexNormal[vertexIndex]=vertexNormal[vertexIndex+1];
						vertexColor[vertexIndex]=vertexColor[vertexIndex+1];
					}
				}

				for(U32 slice=0;slice<slices;slice++)
				{
					const Real angle=2.0*M_PI*slice/(Real)(slices);
					vertexNormal[slice*2+1]=
							SpatialVector(0.0,sin(angle),cos(angle));
					vertex[slice*2+1]=radius*vertexNormal[slice*2+1];
				}
				vertex[slices*2+1]=vertex[1];
				vertexNormal[slices*2+1]=vertexNormal[1];

				if(!copying)
				{
					const Real firstRadius=a_radius? a_radius[0]: 1.0;
					const Real radiusScale=radius>0.0? firstRadius/radius: 1.0;
					for(U32 slice=0;slice<slices;slice++)
					{
						vertexNormal[slice*2]=vertexNormal[slice*2+1];
						vertex[slice*2]=radiusScale*vertex[slice*2+1]-
								SpatialVector(segLength,0.0,0.0);
					}
					vertex[slices*2]=vertex[0];
					vertexNormal[slices*2]=vertexNormal[0];
				}

				for(U32 vertexIndex=copying;vertexIndex<vertexCount;
						vertexIndex+=(1+copying))
				{
					transformVector(segmentFrame,
							vertex[vertexIndex],
							vertex[vertexIndex]);
					rotateVector(segmentFrame,
							vertexNormal[vertexIndex],
							vertexNormal[vertexIndex]);

					vertexColor[vertexIndex]=color;
				}

				a_rspDraw->drawTriangles(vertex,vertexNormal,NULL,vertexCount,
						DrawI::e_strip,TRUE,vertexColor);
			}
		}

		lastPoint=point;
	}
	if(a_curveMode==e_panels)
	{
		a_rspDraw->drawTriangles(vertex,vertexNormal,
				NULL,vertexCount,DrawI::e_strip,
				TRUE,vertexColor);
		a_rspDraw->drawTriangles(vertex2,vertexNormal2,
				NULL,vertexCount,DrawI::e_strip,
				TRUE,vertexColor);
	}
}

//* static
I32 OperateCommon::nearestPointIndex(
	const sp<SurfaceAccessorI> a_spPrimitiveVertices,
	sp<SurfaceCurves::Impact>& a_rspCurveImpact)
{
	const SpatialBary bary=a_rspCurveImpact->barycenter();
	const I32 primitiveIndex=a_rspCurveImpact->primitiveIndex();

	const I32 subCount=a_spPrimitiveVertices->subCount(primitiveIndex);
	const I32 vertexIndex=bary[0]*(subCount-1)+0.5;

	return a_spPrimitiveVertices->integer(primitiveIndex,vertexIndex);
}

//* static
I32 OperateCommon::nearestTriVertexIndex(
	sp<SurfaceTriangles::Impact>& a_rspTriImpact)
{
	const I32 pointIndex0=a_rspTriImpact->pointIndex0();
	const I32 pointIndex1=a_rspTriImpact->pointIndex1();
	const I32 pointIndex2=a_rspTriImpact->pointIndex2();

	const SpatialBary bary=a_rspTriImpact->barycenter();

	SpatialVector bary3=bary;
	bary3[0]=(pointIndex0<0.0)? 0.0: bary3[0];
	bary3[1]=(pointIndex1<0.0)? 0.0: bary3[1];
	bary3[2]=(pointIndex2<0.0)? 0.0: bary3[2];

	if(bary3[0]>bary3[1])
	{
		return (bary3[0]>bary3[2])? 0: 2;
	}
	return (bary3[1]>bary3[2])? 1: 2;
}

//* static
I32 OperateCommon::nearestPointIndex(
	sp<SurfaceTriangles::Impact>& a_rspTriImpact)
{
	const I32 triVertexIndex=nearestTriVertexIndex(a_rspTriImpact);

	if(triVertexIndex==0)
	{
		return a_rspTriImpact->pointIndex0();
	}
	else if (triVertexIndex==1)
	{
		return a_rspTriImpact->pointIndex1();
	}
	return a_rspTriImpact->pointIndex2();
}

//* static
SpatialVector OperateCommon::nearestVertexPoint(
	sp<SurfaceTriangles::Impact>& a_rspTriImpact)
{
	const I32 triVertexIndex=nearestTriVertexIndex(a_rspTriImpact);

	if(triVertexIndex==0)
	{
		return a_rspTriImpact->vertex0();
	}
	else if (triVertexIndex==1)
	{
		return a_rspTriImpact->vertex1();
	}
	return a_rspTriImpact->vertex2();
}


Real OperateCommon::nearestPointAlongCurve(const SpatialVector a_point,
	const sp<SurfaceAccessorI> a_spPrimitiveVertices,
	const I32 a_primitiveIndex)
{
	const U32 subCount=a_spPrimitiveVertices->subCount(a_primitiveIndex);
	if(subCount<2)
	{
		return 0.0;
	}

#if TRUE
	//* TODO return 'unit along' versus 'fractional CV'

	SpatialVector vertexArray[subCount];
	for(U32 subIndex=0;subIndex<subCount;subIndex++)
	{
		vertexArray[subIndex]=a_spPrimitiveVertices->spatialVector(
				a_primitiveIndex,subIndex);
	}

	const Real radius=0.0;
	SpatialVector direction;
	Real along=0.0;
	SpatialVector intersection;

//	const Real distance=
	PointCurveNearest<Real>::solve(vertexArray,subCount,
			radius,a_point,direction,along,intersection);

	return (subCount-1)*along;
#else

	Real result=0.0;
	Real minDist=0.0;

	SpatialVector lastCV=
			a_spPrimitiveVertices->spatialVector(a_primitiveIndex,0);
	for(U32 subIndex=1;subIndex<subCount;subIndex++)
	{
		const SpatialVector cv=a_spPrimitiveVertices->spatialVector(
				a_primitiveIndex,subIndex);
		const SpatialVector segment=cv-lastCV;
		const Real segLength=magnitude(segment);
		const Real safeLength=segLength>0.0? segLength: 1.0;
		const SpatialVector unitSeg=segment/safeLength;
		const SpatialVector away=a_point-lastCV;
		const Real along=dot(unitSeg,away);
		const Real beyond=
				along>segLength? along-segLength: (along<0.0? -along: 0.0);
		const SpatialVector aside=away-along*unitSeg;
		const Real astray=magnitude(aside);
		const Real dist=sqrt(astray*astray+beyond*beyond);

		if(subIndex==1 || dist<minDist)
		{
			const Real unitAlong=along/safeLength;

			minDist=dist;
			result=subIndex-1+unitAlong;
		}

		lastCV=cv;
	}

	return result;
#endif
}

//* static
Real OperateCommon::fractionAtLength(
	const sp<SurfaceAccessorI> a_spPrimitiveVertices,
	const I32 a_primitiveIndex,Real a_length)
{
	const U32 subCount=a_spPrimitiveVertices->subCount(a_primitiveIndex);
	if(subCount<2)
	{
		return 0.0;
	}

	Real cumulative=0.0;
	Real fraction=1.0;

	SpatialVector lastCV=
			a_spPrimitiveVertices->spatialVector(a_primitiveIndex,0);
	U32 subIndex=1;
	for(;subIndex<subCount;subIndex++)
	{
		const SpatialVector cv=a_spPrimitiveVertices->spatialVector(
				a_primitiveIndex,subIndex);
		const SpatialVector segment=cv-lastCV;
		const Real segLength=magnitude(segment);

		cumulative+=segLength;

		if(cumulative>a_length)
		{
			const Real lastLength=cumulative-segLength;
			const Real between=(a_length-lastLength)/segLength;

			fraction=(subIndex-1.0+between)/(subCount-1.0);
			break;
		}

		lastCV=cv;
	}
	if(subIndex==subCount && cumulative>0.0)
	{
		fraction=a_length/cumulative;
	}

//	feLog("OperateCommon::fractionAtLength"
//			" length %.6G fraction %.6G vs %.6G \n",
//			a_length,fraction,a_length/25.0);

	FEASSERT(fraction>=0.0);

	return fraction;
}

//* static
Real OperateCommon::lengthAtFraction(
	const sp<SurfaceAccessorI> a_spPrimitiveVertices,
	const I32 a_primitiveIndex,Real a_fraction)
{
	if(a_fraction<=0.0)
	{
		return 0.0;
	}

	const U32 subCount=a_spPrimitiveVertices->subCount(a_primitiveIndex);
	if(subCount<2)
	{
		return 0.0;
	}

	const Real subCount1=subCount-1.0;
	const Real realCV=a_fraction*subCount1;
	const U32 stopAt=realCV;
	const Real between=realCV-stopAt;

	Real length=0.0;

	SpatialVector lastCV=
			a_spPrimitiveVertices->spatialVector(a_primitiveIndex,0);
	U32 subIndex=1;
	for(;subIndex<subCount;subIndex++)
	{
		const SpatialVector cv=a_spPrimitiveVertices->spatialVector(
				a_primitiveIndex,subIndex);
		const SpatialVector segment=cv-lastCV;
		const Real segLength=magnitude(segment);

		if(subIndex>stopAt)
		{
			length+=between*segLength;
			break;
		}

		length+=segLength;
		lastCV=cv;
	}
	if(subIndex==subCount)
	{
		FEASSERT(a_fraction>0.999);
		length*=a_fraction;
	}

//	feLog("OperateCommon::lengthAtFraction"
//			" fraction %.6G length %.6G vs %.6G \n",
//			a_fraction,length,a_fraction*25.0);
//	feLog("  subCount %d subCount1 %.6G realCV %.6G stopAt %d between %.6G\n",
//			subCount,subCount1,realCV,stopAt,between);

	FEASSERT(length>=0.0);

	return length;
}


SpatialVector OperateCommon::evaluateAtFraction(
	const sp<SurfaceAccessorI> a_spPrimitiveVertices,
	const I32 a_primitiveIndex,Real a_fraction,U32 a_derivitive)
{
	const U32 subCount=a_spPrimitiveVertices->subCount(a_primitiveIndex);
	if(subCount<2)
	{
		return SpatialVector(0.0,0.0,0.0);
	}

	const Real subCount1=subCount-1.0;
	const Real realCV=a_fraction*subCount1;
	const U32 stopAt=realCV;
	const Real between=realCV-stopAt;

	const U32 sub0=stopAt>0? (stopAt<subCount+1? stopAt-1: subCount-1): 0;
	const U32 sub1=stopAt<subCount? stopAt: subCount-1;
	const U32 sub2=stopAt<subCount-1? stopAt+1: subCount-1;
	const U32 sub3=stopAt<subCount-2? stopAt+2: subCount-1;

	const SpatialVector p0=a_spPrimitiveVertices->spatialVector(
			a_primitiveIndex,sub0);
	const SpatialVector p1=a_spPrimitiveVertices->spatialVector(
			a_primitiveIndex,sub1);
	const SpatialVector p2=a_spPrimitiveVertices->spatialVector(
			a_primitiveIndex,sub2);
	const SpatialVector p3=a_spPrimitiveVertices->spatialVector(
			a_primitiveIndex,sub3);

	SpatialVector result;

	switch(a_derivitive)
	{
		case 0:
			result=Spline<SpatialVector,Real>::Cardinal2D(
					between,p0,p1,p2,p3);
			break;
		case 1:
			result=Spline<SpatialVector,Real>::Cardinal2D_d1(
					between,p0,p1,p2,p3);
			break;
		case 2:
			result=Spline<SpatialVector,Real>::Cardinal2D_d2(
					between,p0,p1,p2,p3);
			break;
		default:
			set(result);
			break;
	}

//	feLog("OperateCommon::evaluateAtFraction fraction %.6G deriv %d\n",
//			a_fraction,a_derivitive);
//	feLog("  subCount %d subCount1 %.6G realCV %.6G stopAt %d between %.6G\n",
//			subCount,subCount1,realCV,stopAt,between);
//	feLog("  p0 %s\n",c_print(p0));
//	feLog("  p1 %s\n",c_print(p1));
//	feLog("  p2 %s\n",c_print(p2));
//	feLog("  p3 %s\n",c_print(p3));
//	feLog("  result %s\n",c_print(result));

	return result;
}

//* static
void OperateCommon::drawTubeAxes(sp<DrawI>& a_rspDraw,
	const SpatialTransform& a_transform,Real a_scale,
	Real a_alpha,const Array<Real>& a_profile,
	CurveMode a_curveMode)
{
	const Color rgb[3]=
	{
		Color(1.0,0.7,0.7,a_alpha),
		Color(0.7,1.0,0.7,a_alpha),
		Color(0.7,0.7,1.0,a_alpha)
	};

	const SpatialVector towardCamera(0.0,0.0,1.0);

	const U32 resolution=a_profile.size();

	SpatialVector pointArray[resolution];
	SpatialVector normalArray[resolution];
	Real radiusArray[resolution];
	Color colorArray[resolution];
	Real maxRadius=1e-6;

	for(U32 axis=0;axis<3;axis++)
	{
		for(U32 m=0;m<resolution;m++)
		{
			const Real fraction=m/(resolution-1.0);

			set(pointArray[m]);
			pointArray[m][axis]=a_scale*fraction;
			transformVector(a_transform,pointArray[m],pointArray[m]);

			set(normalArray[m]);
			normalArray[m][(axis+1)%3]=1.0;
			rotateVector(a_transform,normalArray[m],normalArray[m]);

			radiusArray[m]=a_profile[m];
			colorArray[m]=rgb[axis];

			if(maxRadius<radiusArray[m])
			{
				maxRadius=radiusArray[m];
			}
		}

		//* darken narrow areas
		for(U32 m=0;m<resolution;m++)
		{
			colorArray[m]*=pow(radiusArray[m]/maxRadius,0.5);
			colorArray[m][3]=a_alpha;
		}

		const BWORD multicolor=TRUE;
		drawTube(a_rspDraw,pointArray,normalArray,radiusArray,
				resolution,a_curveMode,multicolor,colorArray,towardCamera);
	}
}

//* static
void OperateCommon::drawAnchor(sp<DrawI>& a_rspDraw,
	SpatialTransform a_transform,Real a_scale,U32 a_resolution,Color a_color)
{
	//* TODO args
	const OperateCommon::CurveMode curveMode=OperateCommon::e_tube;
	const SpatialVector towardCamera(0.0,0.0,1.0);

	const Color white(1.0,1.0,1.0,a_color[3]);

	SpatialVector pointArray[a_resolution];
	SpatialVector normalArray[a_resolution];
	Real radiusArray[a_resolution];
	Color colorArray[a_resolution];

	for(U32 m=0;m<a_resolution;m++)
	{
		const Real fraction=m/(a_resolution-1.0);

		set(pointArray[m],0.0,a_scale*fraction,0.0);
		transformVector(a_transform,pointArray[m],pointArray[m]);

		set(normalArray[m],1.0,0.0,0.0);
		rotateVector(a_transform,normalArray[m],normalArray[m]);

		colorArray[m]=a_color*fraction;
		colorArray[m][3]=a_color[3];

		radiusArray[m]=(m==a_resolution-1)?
				0.0: 0.5*a_scale*pow(0.5+0.5*fraction,4.0);
	}

	colorArray[a_resolution-1]=white;

	const BWORD multicolor=TRUE;
	drawTube(a_rspDraw,pointArray,normalArray,radiusArray,
			a_resolution,curveMode,multicolor,colorArray,towardCamera);
}

//* static
void OperateCommon::drawCrossSection(sp<DrawI>& a_rspDraw,
	SpatialVector a_p0,SpatialVector a_p1,SpatialVector a_p2,
	SpatialVector a_planeCenter,SpatialVector a_planeNormal,
	Color a_color)
{
	const SpatialVector diff0=a_p0-a_planeCenter;
	const SpatialVector diff1=a_p1-a_planeCenter;
	const SpatialVector diff2=a_p2-a_planeCenter;

	const Real dot0=dot(diff0,a_planeNormal);
	const Real dot1=dot(diff1,a_planeNormal);
	const Real dot2=dot(diff2,a_planeNormal);

	const BWORD above0=(dot0>=0.0);
	const BWORD above1=(dot1>=0.0);
	const BWORD above2=(dot2>=0.0);

	const I32 aboves=above0+above1+above2;

	if(!aboves || aboves==3)
	{
		//* all above or all below -> no contact
		return;
	}

	SpatialVector pointA;
	SpatialVector pointB;
	SpatialVector pointC;	//* corner

	if(above0!=above1 && above0!=above2)
	{
		pointC=a_p0;

		pointA=a_p1;
		pointB=a_p2;
	}
	else if(above0!=above1 && above1!=above2)
	{
		pointC=a_p1;

		pointA=a_p0;
		pointB=a_p2;
	}
	else
	{
		pointC=a_p2;

		pointA=a_p0;
		pointB=a_p1;
	}

	const SpatialVector unitCA=unitSafe(pointA-pointC);
	const SpatialVector unitCB=unitSafe(pointB-pointC);

	const Real distCA=RayPlaneIntersect<Real>::solve(
			a_planeCenter,a_planeNormal,-1.0,pointC,unitCA);
	const Real distCB=RayPlaneIntersect<Real>::solve(
			a_planeCenter,a_planeNormal,-1.0,pointC,unitCB);

	SpatialVector line[2];
	line[0]=pointC+distCA*unitCA;
	line[1]=pointC+distCB*unitCB;

	a_rspDraw->drawLines(line,NULL,2,DrawI::e_strip,false,&a_color);
}

void OperateCommon::drawCrossSection(sp<DrawI>& a_rspDraw,
	sp<SurfaceAccessorI> a_spVertices,
	SpatialVector a_planeCenter,
	SpatialVector a_planeNormal,
	Color a_color)
{
	const U32 primitiveCount=a_spVertices->count();
	for(U32 primitiveIndex=0;primitiveIndex<primitiveCount;
			primitiveIndex++)
	{
		const U32 subCount=a_spVertices->subCount(primitiveIndex);
		if(subCount<3 || subCount>4)
		{
			continue;
		}

		const SpatialVector p0=
				a_spVertices->spatialVector(primitiveIndex,0);
		const SpatialVector p1=
				a_spVertices->spatialVector(primitiveIndex,1);
		const SpatialVector p2=
				a_spVertices->spatialVector(primitiveIndex,2);

		OperateCommon::drawCrossSection(a_rspDraw,p0,p1,p2,
				a_planeCenter,a_planeNormal,a_color);

		if(subCount==3)
		{
			continue;
		}

		const SpatialVector p3=
				a_spVertices->spatialVector(primitiveIndex,3);

		OperateCommon::drawCrossSection(a_rspDraw,p1,p2,p3,
				a_planeCenter,a_planeNormal,a_color);
	}
}

//* static
Real OperateCommon::flameProfile(Real a_peakOffset,Real a_fraction)
{
	const Real maxOffset=0.9*a_fraction;	//* tweak
	Real offset=a_peakOffset*sin(M_PI*a_fraction);
	if(offset>maxOffset)
	{
		offset=maxOffset;
	}
	return sin(M_PI*(a_fraction-offset));
}

