/*	Copyright (C) 2003-2015 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#ifndef __operator_FollicleOp_h__
#define __operator_FollicleOp_h__

namespace fe
{

/**************************************************************************//**
	@brief Operator to edit surface locations

	@ingroup operator
*//***************************************************************************/
class FE_DL_EXPORT FollicleOp:
	public OperatorState,
	public Initialize<FollicleOp>
{
	public:

						FollicleOp(void):
							m_brushed(FALSE),
							m_lastX(-1),
							m_lastY(-1),
							m_lastUV(-1,-1),
							m_dragging(FALSE),
							m_areaSelect(FALSE),
							m_addRadius(0.0),
							m_addFalloff(0.0),
							m_addWheel(0)									{}

virtual					~FollicleOp(void)									{}

		void			initialize(void);

						//* As HandlerI
virtual	void			handle(Record& a_rSignal);

	protected:

		void			syncToInput(void);

		void			addFollicle(const SpatialVector& a_intersection);

		Record			getEdit(I32 a_index);
		void			movePicked(const SpatialVector& a_rDelta);
		void			slideOnUV(I32 a_index,const Vector2 a_deltaUV);
		void			snapToSurface(I32 a_index);

		void			setValueLimited(Record a_edit,SpatialVector a_value);
		void			addValue(SpatialVector a_increase,
								SpatialVector a_mask,BWORD a_flood);
		void			scaleValue(Real a_scale,
								SpatialVector a_mask,BWORD a_flood);
		void			paintValue(SpatialVector a_replacement,
								SpatialVector a_mask,BWORD a_flood);
		void			restoreValue(SpatialVector a_mask,BWORD a_flood);

		BWORD			calcWeightedAverage(SpatialVector& rAverage,
								BWORD a_flood);
		BWORD			calcLimits(SpatialVector& rMinimum,
								SpatialVector& rMaximum,BWORD a_flood);
		void			smoothValue(SpatialVector a_mask,BWORD a_flood);
		void			convergeValue(SpatialVector a_mask,BWORD a_flood);
		void			divergeValue(SpatialVector a_mask,BWORD a_flood);

		BWORD			revert(I32 a_index);
		BWORD			hideToggle(I32 a_index);
		BWORD			hidden(I32 a_index);
		void			hide(I32 a_index);

		BWORD			moved(I32 a_index);

virtual	void			setupState(void);
virtual	BWORD			loadState(const String& rBuffer);

	private:

		void			updateAction(String& m_rAction);
		void			updateNeighbors(void);

		SpatialVector	inputValue(I32 a_pointIndex);

		sp<SurfaceAccessorI>		m_spInputPoint;
		sp<SurfaceAccessorI>		m_spOutputPoint;
		sp<SurfaceAccessorI>		m_spOutputNormal;
		sp<SurfaceAccessorI>		m_spOutputUV;
		sp<SurfaceAccessorI>		m_spInputAttribute;
		sp<SurfaceAccessorI>		m_spOutputAttribute;
		sp<SurfaceAccessorI>		m_spHideGroup;
		sp<SurfaceI>				m_spDriver;
		sp<SurfaceI>				m_spPoints;
		WindowEvent					m_event;
		String						m_attributeType;

		sp<DrawMode>				m_spWireframe;
		sp<DrawMode>				m_spSolid;
		sp<DrawMode>				m_spForeshadow;
		sp<DrawMode>				m_spBehind;

		sp< RecordMap<I32> >		m_spEditMap;

		Accessor<SpatialVector>		m_aSyncLocation;
		Accessor<SpatialVector>		m_aLocation;
		Accessor<Vector2>			m_aUV;
		Accessor<SpatialVector>		m_aValue;

		BWORD						m_brushed;
		I32							m_lastX;
		I32							m_lastY;
		Vector2						m_lastUV;
		SpatialVector				m_pickPoint;
		SpatialVector				m_pickNormal;

		Array< Array<I32> >			m_neighborTable;
		Array<Color>				m_colorRim;
		Array<Color>				m_colorCenter;
		BWORD						m_dragging;
		BWORD						m_areaSelect;
		Real						m_addRadius;
		Real						m_addFalloff;
		I32							m_addWheel;
};

} /* namespace */

#endif /* __operator_FollicleOp_h__ */
