/*	Copyright (C) 2003-2015 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#define FE_HGO_DEBUG		FALSE
#define FE_HGO_VERBOSE		FALSE
#define FE_HGO_PROFILE		FALSE

#include "operator/operator.pmh"

using namespace fe;

void HingeOp::initialize(void)
{
	catalog<String>("icon")="FE_alpha";

	catalog<String>("Threads","page")="Config";

	catalog<String>("AutoThread","page")="Config";

	catalog<String>("Paging","page")="Config";

	catalog<String>("StayAlive","page")="Config";

	catalog<String>("Work","page")="Config";

	//* page Config

	catalog<bool>("fragment")=false;
	catalog<String>("fragment","label")="Fragment";
	catalog<String>("fragment","page")="Config";
	catalog<bool>("fragment","joined")=true;
	catalog<String>("fragment","hint")=
			"Associate collections of primitives which have"
			" matching a fragment attribute.";

	catalog<String>("fragmentAttr")="part";
	catalog<String>("fragmentAttr","label")="By Attr";
	catalog<String>("fragmentAttr","page")="Config";
	catalog<String>("fragmentAttr","enabler")="fragment";
	catalog<bool>("fragmentAttr","joined")=true;
	catalog<String>("fragmentAttr","hint")=
			"Name of string attribute describing membership"
			" in a collection of primitives.";

	catalog<bool>("fragmentCollide")=false;
	catalog<String>("fragmentCollide","label")="Inter-Collide";
	catalog<String>("fragmentCollide","page")="Config";
	catalog<String>("fragmentCollide","enabler")="fragment";
	catalog<String>("fragment","hint")=
			"Check for collisions with other fragments.";

	catalog<String>("pivotMethod")="manual";
	catalog<String>("pivotMethod","label")="Pivot Method";
	catalog<String>("pivotMethod","choice:0")="manual";
	catalog<String>("pivotMethod","label:0")="Manual";
	catalog<String>("pivotMethod","choice:1")="du";
	catalog<String>("pivotMethod","label:1")="About Driver Du";
	catalog<bool>("pivotMethod","joined")=true;
	catalog<String>("pivotMethod","page")="Config";

	catalog<bool>("reversePivot")=false;
	catalog<String>("reversePivot","label")="Reverse";
	catalog<String>("reversePivot","page")="Config";

	catalog<I32>("faceA")=0;
	catalog<I32>("faceA","high")=100;
	catalog<I32>("faceA","max")=INT_MAX;
	catalog<String>("faceA","label")="Face A";
	catalog<bool>("faceA","joined")=true;
	catalog<String>("faceA","page")="Config";

	catalog<SpatialVector>("baryA")=SpatialVector(1.0,0.0,0.0);
	catalog<String>("baryA","label")="Barycenter A";
	catalog<String>("baryA","page")="Config";

	catalog<SpatialVector>("offsetA");
	catalog<String>("offsetA","label")="Offset A";
	catalog<String>("offsetA","page")="Config";

	catalog<I32>("faceB")=0;
	catalog<I32>("faceB","high")=100;
	catalog<I32>("faceB","max")=INT_MAX;
	catalog<String>("faceB","label")="Face B";
	catalog<bool>("faceB","joined")=true;
	catalog<String>("faceB","page")="Config";

	catalog<SpatialVector>("baryB")=SpatialVector(1.0,0.0,0.0);
	catalog<String>("baryB","label")="Barycenter B";
	catalog<String>("baryB","page")="Config";

	catalog<SpatialVector>("offsetB");
	catalog<String>("offsetB","label")="Offset B";
	catalog<String>("offsetB","page")="Config";

	catalog<I32>("faceC")=0;
	catalog<I32>("faceC","high")=100;
	catalog<I32>("faceC","max")=INT_MAX;
	catalog<String>("faceC","label")="Face C";
	catalog<bool>("faceC","joined")=true;
	catalog<String>("faceC","page")="Config";

	catalog<SpatialVector>("baryC")=SpatialVector(1.0,0.0,0.0);
	catalog<String>("baryC","label")="Barycenter C";
	catalog<String>("baryC","page")="Config";

	catalog<SpatialVector>("offsetC");
	catalog<String>("offsetC","label")="Offset C";
	catalog<String>("offsetC","page")="Config";

	//* page Behavior

	catalog<String>("collideMethod")="ray";
	catalog<String>("collideMethod","label")="Collision Method";
	catalog<String>("collideMethod","choice:0")="ray";
	catalog<String>("collideMethod","label:0")="Ray Cast";
	catalog<String>("collideMethod","choice:1")="nearest";
	catalog<String>("collideMethod","label:1")="Nearest Point";
	catalog<String>("collideMethod","page")="Behavior";
	catalog<bool>("collideMethod","joined")=true;

	catalog<I32>("hitLimit")=8;
	catalog<String>("hitLimit","label")="Hit Limit";
	catalog<I32>("hitLimit","min")=1;
	catalog<I32>("hitLimit","high")=16;
	catalog<I32>("hitLimit","max")=100;
	catalog<String>("hitLimit","page")="Behavior";

	catalog<bool>("Temporal")=false;
	catalog<String>("Temporal","page")="Behavior";
	catalog<String>("Temporal","hint")=
			"Tries to retain state of the prior frame."
			"  If the prior frame was not the last evaluated,"
			" the node may choose not to use temporal data.";
	catalog<bool>("Temporal","joined")=true;

	catalog<Real>("Response")=1.0;
	catalog<String>("Response","enabler")="Temporal";
	catalog<String>("Response","page")="Behavior";
	catalog<String>("Response","hint")=
			"When temporal, this defines how quickly the surface returns"
			" to its original shaped when released.";

	catalog<Real>("angleLimit")=180.0;
	catalog<Real>("angleLimit","high")=360.0;
	catalog<String>("angleLimit","label")="Angle Limit";
	catalog<String>("angleLimit","page")="Behavior";
	catalog<String>("angleLimit","hint")=
			"Stops the axial rotation from exceeding this limit.";

	catalog<Real>("deadRadius")=0.0;
	catalog<Real>("deadRadius","high")=10.0;
	catalog<Real>("deadRadius","max")=1e6;
	catalog<String>("deadRadius","label")="Dead Radius";
	catalog<bool>("deadRadius","joined")=true;
	catalog<String>("deadRadius","page")="Behavior";

	catalog<Real>("liveRadius")=0.0;
	catalog<Real>("liveRadius","high")=10.0;
	catalog<Real>("liveRadius","max")=1e6;
	catalog<String>("liveRadius","label")="Live Radius";
	catalog<String>("liveRadius","page")="Behavior";

	catalog<Real>("threshold")=0.0;
	catalog<Real>("threshold","min")= -1e6;
	catalog<Real>("threshold","low")=0.0;
	catalog<Real>("threshold","high")=1.0;
	catalog<Real>("threshold","max")=1e6;
	catalog<String>("threshold","label")="Threshold";
	catalog<String>("threshold","page")="Behavior";

	catalog<bool>("twist")=false;
	catalog<String>("twist","label")="Twist";
	catalog<bool>("twist","joined")=true;
	catalog<String>("twist","page")="Behavior";

	catalog<Real>("twistiness")=1.0;
	catalog<Real>("twistiness","high")=1.0;
	catalog<Real>("twistiness","max")=1e6;
	catalog<String>("twistiness","label")="Twistiness";
	catalog<String>("twistiness","page")="Behavior";

	catalog<bool>("soft")=false;
	catalog<String>("soft","label")="Soft";
	catalog<bool>("soft","joined")=true;
	catalog<String>("soft","page")="Behavior";

	catalog<Real>("softness")=1.0;
	catalog<Real>("softness","high")=1.0;
	catalog<Real>("softness","max")=1e6;
	catalog<String>("softness","label")="Softness";
	catalog<String>("softness","page")="Behavior";

	catalog<I32>("passes")=1;
	catalog<I32>("passes","min")=1;
	catalog<I32>("passes","high")=10;
	catalog<I32>("passes","max")=1000;
	catalog<String>("passes","label")="Passes";
	catalog<String>("passes","page")="Behavior";
	catalog<bool>("passes","joined")=true;

	catalog<Real>("rate")=1.0;
	catalog<Real>("rate","max")=1.0;
	catalog<String>("rate","label")="Rate";
	catalog<String>("rate","page")="Behavior";

	//* page Display

	catalog<bool>("isolate")=false;
	catalog<String>("isolate","label")="Isolate";
	catalog<bool>("isolate","joined")=true;
	catalog<String>("isolate","page")="Display";

	catalog<String>("isolateFragment");
	catalog<String>("isolateFragment","label")="Fragment";
	catalog<String>("isolateFragment","page")="Display";

	catalog< sp<Component> >("Input Surface");

	catalog< sp<Component> >("Collider Surface");

	catalog< sp<Component> >("Driver Surface");
	catalog<bool>("Driver Surface","optional")=true;

	catalog< sp<Component> >("Output Surface");
	catalog<String>("Output Surface","IO")="output";
	catalog<String>("Output Surface","copy")="Input Surface";
	catalog<bool>("Output Surface","recycle")=true;
	catalog<bool>("Output Surface","temporal")=true;
}

void HingeOp::handle(Record& a_rSignal)
{
#if FE_HGO_DEBUG
	feLog("HingeOp::handle \"%s\"\n",name().c_str());
#endif

#if FE_HGO_PROFILE
	m_spProfiler=new Profiler("HingeOp");
	m_spProfilePrep=new Profiler::Profile(m_spProfiler,"Prep");
	m_spProfileLoop=new Profiler::Profile(m_spProfiler,"Loop");
	m_spProfilePush=new Profiler::Profile(m_spProfiler,"Push");
	m_spProfilePull=new Profiler::Profile(m_spProfiler,"Pull");
	m_spProfileRotate=new Profiler::Profile(m_spProfiler,"Rotate");

	m_spProfiler->begin();
	m_spProfilePrep->start();
#endif

	String& rSummary=catalog<String>("summary");
	rSummary="";

	const String fragmentAttr=catalog<String>("fragmentAttr");
	const BWORD fragmentCollide=catalog<bool>("fragmentCollide");
	const String pivotMethod=catalog<String>("pivotMethod");
	const BWORD reversePivot=catalog<bool>("reversePivot");
	const Real twistiness=catalog<Real>("twistiness");
	const BWORD twist=(twistiness>0.0 && catalog<bool>("twist"));
	const Real softness=catalog<Real>("softness");
	const BWORD soft=(softness>0.0 && catalog<bool>("soft"));
	const I32 passes=catalog<I32>("passes");
	const Real rate=catalog<Real>("rate");
	const BWORD isolate=catalog<bool>("isolate");
	const String isolateFragment=catalog<String>("isolateFragment");

	BWORD fragmented=(!fragmentAttr.empty() && catalog<bool>("fragment"));

	if(!accessOutput(m_spOutputAccessible,a_rSignal)) return;

	sp<SurfaceAccessibleI> spInputAccessible;
	if(!access(spInputAccessible,"Input Surface")) return;

	sp<SurfaceAccessorI> spInputPoint;
	if(!access(spInputPoint,spInputAccessible,e_point,e_position)) return;

	sp<SurfaceAccessorI> spOutputPoint;
	if(!access(spOutputPoint,m_spOutputAccessible,e_point,e_position)) return;

	sp<SurfaceAccessorI> spOutputVertices;
	if(!access(spOutputVertices,m_spOutputAccessible,
			e_primitive,e_vertices)) return;

	sp<SurfaceI> spCollider;
	if(!access(spCollider,"Collider Surface")) return;

	spCollider->setRefinement(0);
	spCollider->prepareForSearch();

	sp<DrawI> spDrawGuide;
	accessGuide(spDrawGuide,a_rSignal,e_warning);

	sp<SurfaceI> spDriver;
	access(spDriver,"Driver Surface",e_quiet);

	m_pivotA=catalog<SpatialVector>("offsetA");
	m_pivotB=catalog<SpatialVector>("offsetB");
	m_pivotC=catalog<SpatialVector>("offsetC");

	const U32 pointCount=spOutputPoint->count();

	if(spDriver.isValid() && pivotMethod=="manual")
	{
		const I32 triangleIndexA=catalog<I32>("faceA");
		const SpatialBary baryA=catalog<SpatialVector>("baryA");
		const SpatialTransform xformA=
				spDriver->sample(triangleIndexA,baryA);
		transformVector(xformA,m_pivotA,m_pivotA);

		const I32 triangleIndexB=catalog<I32>("faceB");
		const SpatialBary baryB=catalog<SpatialVector>("baryB");
		const SpatialTransform xformB=
				spDriver->sample(triangleIndexB,baryB);
		transformVector(xformB,m_pivotB,m_pivotB);

		const I32 triangleIndexC=catalog<I32>("faceC");
		const SpatialBary baryC=catalog<SpatialVector>("baryC");
		const SpatialTransform xformC=
				spDriver->sample(triangleIndexC,baryC);
		transformVector(xformC,m_pivotC,m_pivotC);
	}

	if(reversePivot)
	{
		const SpatialVector store=m_pivotA;
		m_pivotA=m_pivotB;
		m_pivotB=store;
	}

	const BWORD temporal=catalog<bool>("Temporal");
	const Real response=catalog<Real>("Response");
	const Real radianLimit=catalog<Real>("angleLimit")*degToRad;

	const BWORD doTemporal=(temporal && response<1.0);

	if(doTemporal)
	{
		rSummary+=String(rSummary.empty()? "": " + ")+"temporal";
	}
	if(twist)
	{
		rSummary+=String(rSummary.empty()? "": " + ")+"twist";
	}
	if(soft)
	{
		rSummary+=String(rSummary.empty()? "": " + ")+"soft";
	}

	const I32 frame=currentFrame(a_rSignal);
	const BWORD outputReplaced=catalog<bool>("Output Surface","replaced");
	const BWORD useCache=doTemporal && m_pointAngle.size()==pointCount &&
			(frame==m_frame+1 || (frame==m_frame && !outputReplaced));

#if FE_HGO_DEBUG
	feLog("HingeOp::handle frame %d->%d outputReplaced %d"
			" pointCount %d cache %d\n",
			m_frame,frame,outputReplaced,pointCount,m_pointAngle.size());
#endif

	if(!outputReplaced)
	{
		for(U32 pointIndex=0;pointIndex<pointCount;pointIndex++)
		{
			const SpatialVector point=
					spInputPoint->spatialVector(pointIndex);
			spOutputPoint->set(pointIndex,point);
		}
	}

	//* presuming always replaced with frame change
	if(useCache)
	{
#if FE_HGO_DEBUG
		if(frame==m_frame+1)
		{
			feLog("HingeOp::handle temporal step\n");
		}
		else
		{
			feLog("HingeOp::handle temporal repeat\n");
		}
#endif

		FEASSERT(m_pointAngle.size()==pointCount);
		FEASSERT(m_totalAngle.size()==pointCount);

		//* output = history:prior blended with pure input
		for(U32 pointIndex=0;pointIndex<pointCount;pointIndex++)
		{
			m_pointAngle[pointIndex]=m_totalAngle[pointIndex]*(1.0-response);
			m_totalAngle[pointIndex]=0.0;

			if(m_pointAngle[pointIndex]>radianLimit)
			{
				m_pointAngle[pointIndex]=radianLimit;
			}
		}
	}
	else
	{
		m_pointAngle.resize(pointCount);
		m_totalAngle.resize(pointCount);

#if FE_HGO_DEBUG
		feLog("HingeOp::handle rebuild\n");
#endif

		//* output = pure input
		for(U32 pointIndex=0;pointIndex<pointCount;pointIndex++)
		{
			m_pointAngle[pointIndex]=0.0;
			m_totalAngle[pointIndex]=0.0;
		}
	}
	m_frame=frame;

#if FE_HGO_PROFILE
	m_spProfileLoop->replace(m_spProfilePrep);
#endif

	const U32 primitiveCount=spOutputVertices->count();

	sp<SurfaceAccessorI> spInputVertices;
	if(!access(spInputVertices,spInputAccessible,
			e_primitive,e_vertices)) return;

	Array<SpatialVector> fragPivotA;
	Array<SpatialVector> fragPivotB;
	Array<SpatialVector> fragPivotC;

	I32 fragmentCount=1;
	if(fragmented)
	{
		const String inputGroup="";	//* TODO

		spInputVertices->fragmentWith(SurfaceAccessibleI::e_primitive,
				fragmentAttr,inputGroup);
		fragmentCount=spInputVertices->fragmentCount();
		if(!fragmentCount)
		{
			fragmentCount=1;
			fragmented=FALSE;
		}
	}

	for(I32 pass=0;pass<passes;pass++)
	{
		sp<SurfaceI> spOutput;
		if(fragmented && fragmentCollide)
		{
			access(spOutput,m_spOutputAccessible);
			spOutput->partitionWith(fragmentAttr);
			spOutput->setRefinement(0);
			spOutput->prepareForSearch();
		}

#if FE_HGO_DEBUG
		feLog("HingeOp::handle fragmentCount %d\n",fragmentCount);
#endif

		const U32 colliderCount=spOutput.isValid()? 2: 1;
		for(U32 colliderIndex=0;colliderIndex<colliderCount;colliderIndex++)
		{
			m_spObstacle=(colliderIndex==colliderCount-1)? spCollider: spOutput;

			sp<SurfaceAccessibleI::FilterI> spFilter;

			for(I32 fragmentIndex=0;fragmentIndex<fragmentCount;fragmentIndex++)
			{
				m_partitionIndex= -1;

				BWORD doDraw=TRUE;

				I32 filterCount=primitiveCount;
				if(fragmented)
				{
					const String fragmentName=
							spInputVertices->fragment(fragmentIndex);

					if(isolate && fragmentName!=isolateFragment)
					{
						doDraw=FALSE;
					}

					if(!spInputVertices->filterWith(fragmentName,spFilter) ||
							spFilter.isNull())
					{
						continue;
					}

					filterCount=spFilter->filterCount();
					if(!filterCount)
					{
						continue;
					}

					if(spOutput.isValid())
					{
						//* pattern is anything but (negative lookbehind)
						const String pattern=".*(?<!"+fragmentName+")$";
						spOutput->setPartitionFilter(pattern);

						//* TODO confirm partition indexes occur in order
						m_partitionIndex=fragmentName.integer();
					}

					sp<SpannedRange> spSerialRange=
							sp<SpannedRange>(new SpannedRange());

					for(I32 filterIndex=0;filterIndex<filterCount;filterIndex++)
					{
						const I32 reIndex=spFilter->filter(filterIndex);
						const U32 subCount=spInputVertices->subCount(reIndex);

						for(U32 subIndex=0;subIndex<subCount;subIndex++)
						{
							const I32 pointIndex=
									spInputVertices->integer(reIndex,subIndex);
							spSerialRange->nonAtomic().add(
									pointIndex,pointIndex);
						}
					}

					setFullRange(spSerialRange);
				}
				else
				{
					setAtomicRange(m_spOutputAccessible,e_pointsOnly);
				}

#if FE_HGO_PROFILE
				m_spProfilePush->replace(m_spProfileLoop);
#endif

#if FE_HGO_DEBUG
				feLog("HingeOp::handle pass %d/%d collider %d/%d"
						" fragment %d/%d range %s\n",
						pass,passes,
						colliderIndex,colliderCount,
						fragmentIndex,fragmentCount,
						c_print(fullRange()->brief()));
#endif

				if(spDriver.isValid() && pivotMethod=="du")
				{
					if(!pass && !colliderIndex)
					{
						SpatialVector midpoint(0.0,0.0,0.0);

						SpatialVector farPoint=midpoint;
						Real farDistance=0.0;

						for(I32 filterIndex=0;filterIndex<filterCount;
								filterIndex++)
						{
							const I32 reIndex=spFilter->filter(filterIndex);
							const U32 subCount=
									spInputVertices->subCount(reIndex);

							if(!filterIndex && subCount)
							{
								const I32 pointIndex=
										spInputVertices->integer(reIndex,0);
								midpoint=spOutputPoint->spatialVector(
										pointIndex);
							}

							for(U32 subIndex=(filterIndex==0);
									subIndex<subCount;subIndex++)
							{
								const I32 pointIndex=spInputVertices->integer(
										reIndex,subIndex);

								const SpatialVector otherPoint=
										spOutputPoint->spatialVector(
										pointIndex);
								const Real distance=
										magnitude(otherPoint-midpoint);
								if(farDistance<distance)
								{
									farDistance=distance;
									farPoint=otherPoint;
								}
							}
						}

						sp<SurfaceI::ImpactI> spImpact=
								spDriver->nearestPoint(midpoint);
						if(spImpact.isValid())
						{
							// HACK negative (should work with reversePivot)
							const SpatialVector du=spImpact->du();

							m_pivotA=midpoint-du;
							m_pivotB=midpoint+du;
							m_pivotC=farPoint;

							if(reversePivot)
							{
								const SpatialVector store=m_pivotA;
								m_pivotA=m_pivotB;
								m_pivotB=store;
							}
						}

						if(I32(fragPivotA.size())<=fragmentIndex)
						{
							fragPivotA.resize(fragmentIndex+1);
							fragPivotB.resize(fragmentIndex+1);
							fragPivotC.resize(fragmentIndex+1);
						}
						fragPivotA[fragmentIndex]=m_pivotA;
						fragPivotB[fragmentIndex]=m_pivotB;
						fragPivotC[fragmentIndex]=m_pivotC;
					}
					else
					{
						m_pivotA=fragPivotA[fragmentIndex];
						m_pivotB=fragPivotB[fragmentIndex];
						m_pivotC=fragPivotC[fragmentIndex];
					}
				}

				const SpatialVector midpoint=0.5*(m_pivotA+m_pivotB);
				const SpatialVector axis=unitSafe(m_pivotB-m_pivotA);
				const SpatialVector crossAxis=unitSafe(m_pivotC-midpoint);

				if(!pass && !colliderIndex && doDraw && spDrawGuide.isValid())
				{
//						sp<DrawableI> spDrawable(m_spObstacle);
//						if(spDrawable.isValid())
//						{
//							spDrawable->draw(spDrawGuide,NULL);
//						}

					const Color red(0.0,1.0,0.0);
					const Color green(0.0,1.0,0.0);
					const Color blue(0.0,0.0,1.0);

					Color colors[2];
					colors[0]=green;
					colors[1]=blue;

					SpatialVector line[2];
					line[0]=m_pivotA;
					line[1]=m_pivotB;

					spDrawGuide->drawLines(line,NULL,2,DrawI::e_strip,
							TRUE,colors);

					if(twist)
					{
						colors[0]=green;
						colors[1]=red;

						line[0]=midpoint;
						line[1]=m_pivotC;

						spDrawGuide->drawLines(line,NULL,2,DrawI::e_strip,
								TRUE,colors);
					}
				}


				adjustThreads();
				const I32 threadCount=catalog<I32>("Threads");

				m_maxAngle.resize(threadCount);
				for(I32 threadIndex=0;threadIndex<threadCount;threadIndex++)
				{
					m_maxAngle[threadIndex]=0.0;
				}

				if(!fragmented && threadCount>1)
				{
					OperatorThreaded::handle(a_rSignal);
				}
				else
				{
//					m_spDrawGuide=spDrawGuide;
					//* serial
					run(-1,fullRange());
				}

				m_spDrawGuide=NULL;

#if FE_HGO_PROFILE
				m_spProfilePull->replace(m_spProfilePush);
#endif

				if(soft)
				{
					while(TRUE)
					{
						BWORD changed=FALSE;

						for(I32 filterIndex=0;filterIndex<filterCount;
								filterIndex++)
						{
							if(interrupted())
							{
								changed=FALSE;
								break;
							}

							const U32 primitiveIndex=fragmented?
									spFilter->filter(filterIndex): filterIndex;

							const U32 subCount=
									spOutputVertices->subCount(primitiveIndex);
							for(U32 subIndex=0;subIndex<subCount;subIndex++)
							{
								//* TODO cache lookups

								const U32 subIndexB=(subIndex+1)%subCount;

								const U32 pointIndexA=
										spOutputVertices->integer(
										primitiveIndex,subIndex);
								const U32 pointIndexB=
										spOutputVertices->integer(
										primitiveIndex,subIndexB);

								Real& rAngleA=m_pointAngle[pointIndexA];
								Real& rAngleB=m_pointAngle[pointIndexB];

								if(rAngleA<1e-6 && rAngleB<1e-6)
								{
									continue;
								}

								const SpatialVector pointA=
										spInputPoint->spatialVector(
										pointIndexA);
								const SpatialVector pointB=
										spInputPoint->spatialVector(
										pointIndexB);

								const Real distance=magnitude(pointB-pointA);
								if(distance<1e-6)
								{
									continue;
								}

								const Real limit=softness*distance;

//								feLog("vert %d:%d point %d %d angle %.6G %.6G"
//										" limit %.6G\n",
//										primitiveIndex,subIndex,
//										pointIndexA,pointIndexB,
//										rAngleA,rAngleB,limit);

//								const Real minA=rAngleB-limit;
								const Real minA=rAngleB*(1.0-limit);
								if(rAngleA<minA)
								{
									rAngleA=minA;
									changed=TRUE;
								}
								else
								{
//									const Real minB=rAngleA-limit;
									const Real minB=rAngleA*(1.0-limit);
									if(rAngleB<minB)
									{
										rAngleB=minB;
										changed=TRUE;
									}
								}
							}
						}

						if(!changed)
						{
							break;
						}
					}
				}

#if FE_HGO_PROFILE
				m_spProfileRotate->replace(m_spProfilePull);
#endif

				Real maxAngle=0.0;
				const U32 maxAngleCount=m_maxAngle.size();
				for(U32 maxAngleIndex=0;maxAngleIndex<maxAngleCount;
						maxAngleIndex++)
				{
					Real& rMaxAngle=m_maxAngle[maxAngleIndex];

#if FE_HGO_VERBOSE
					feLog("pass %d/%d maxAngle %d/%d %.6G\n",
							pass,passes,maxAngleIndex,maxAngleCount,rMaxAngle);
#endif

					if(maxAngle<rMaxAngle)
					{
						maxAngle=rMaxAngle;
					}
					rMaxAngle=0.0;
				}

				if(maxAngle>0.0)
				{
					SpatialTransform conversion;
					SpatialTransform twistTransform;

					if(!soft)
					{
						calcConversion(conversion,rate*maxAngle,axis,midpoint);
					}

					if(twist)
					{
						Real maxArm=0;
						Real spinMin=0.0;
						Real spinMax=0.0;

						for(SpannedRange::Iterator it=fullRange()->begin();
								!it.atEnd();it.step())
						{
							const U32 pointIndex=it.value();

							const SpatialVector point=
									spOutputPoint->spatialVector(pointIndex);

							const SpatialVector toPoint=point-midpoint;
							const Real alongCrossAxis=dot(crossAxis,toPoint);

							const SpatialVector onCrossAxis=
									midpoint+alongCrossAxis*crossAxis;

							const SpatialVector arm=point-onCrossAxis;
							const Real length=magnitude(arm);
							if(length<1e-6)
							{
								continue;
							}

							const Real aligned=dot(axis,arm);
							const Real absAligned=fabs(aligned);

							if(maxArm<absAligned)
							{
								maxArm=absAligned;
							}

							SpatialTransform oneConversion;
							calcConversion(oneConversion,
									rate*m_pointAngle[pointIndex],
									axis,midpoint);

							SpatialVector transformed;
							transformVector(oneConversion,point,transformed);

							const SpatialVector arm2=transformed-onCrossAxis;

							SpatialQuaternion rotation;
							set(rotation,unitSafe(arm),unitSafe(arm2));

							Real radians;
							SpatialVector rotAxis;
							rotation.computeAngleAxis(radians,rotAxis);

							const Real spin= -aligned*radians;

//							feLog("angle %.6G axis %s arm %s arm2 %s"
//									" radians %.6G spin %.6G\n",
//									m_pointAngle[pointIndex],c_print(axis),
//									c_print(arm),c_print(arm2),radians,spin);

							spinMin=fe::minimum(spinMin,spin);
							spinMax=fe::maximum(spinMax,spin);
						}

						if(maxArm>0.0)
						{
							const Real angle=
									twistiness*(spinMax+spinMin)/maxArm;

//							feLog("spinMin %.6G spinMax %.6G"
//									" maxArm %.6G angle %.6G\n",
//									spinMin,spinMax,maxArm,angle/degToRad);

							const SpatialVector yAxis(0.0,1.0,0.0);
							const SpatialQuaternion rotation(angle,yAxis);

							SpatialTransform xformMidpoint;
							makeFrameNormalY(xformMidpoint,
									midpoint,axis,crossAxis);

							SpatialTransform invMidpoint;
							invert(invMidpoint,xformMidpoint);

							twistTransform=invMidpoint*
									SpatialTransform(rotation)*xformMidpoint;

							if(!soft)
							{
								conversion*=twistTransform;
							}
						}
					}

#if FE_HGO_VERBOSE
					feLog("conversion\n%s\n",c_print(conversion));
#endif

					//* rotate all points
					for(SpannedRange::Iterator it=fullRange()->begin();
							!it.atEnd();it.step())
					{
						const U32 pointIndex=it.value();

						if(soft)
						{
							calcConversion(conversion,
									rate*m_pointAngle[pointIndex],
									axis,midpoint);

							if(twist)
							{
								conversion*=twistTransform;
							}
						}

						const SpatialVector point=
								spOutputPoint->spatialVector(pointIndex);

						SpatialVector transformed;
						transformVector(conversion,point,transformed);

						spOutputPoint->set(pointIndex,transformed);
					}
				}

				for(SpannedRange::Iterator it=fullRange()->begin();
						!it.atEnd();it.step())
				{
					const U32 pointIndex=it.value();

					m_totalAngle[pointIndex]+=m_pointAngle[pointIndex];
					m_pointAngle[pointIndex]=0.0;
				}

#if FE_HGO_PROFILE
				m_spProfileLoop->replace(m_spProfileRotate);
#endif
			}
		}
	}

	if(fragmented && isolate)
	{
		sp<SurfaceAccessibleI::FilterI> spFilter;

		for(I32 fragmentIndex=0;fragmentIndex<fragmentCount;fragmentIndex++)
		{
			const String fragmentName=spInputVertices->fragment(fragmentIndex);
			if(fragmentName==isolateFragment)
			{
				continue;
			}

			I32 filterCount=primitiveCount;

			if(!spInputVertices->filterWith(fragmentName,spFilter) ||
					spFilter.isNull())
			{
				continue;
			}

			filterCount=spFilter->filterCount();
			if(!filterCount)
			{
				continue;
			}

			std::set<I32> pointSet;

			const I32 rootIndex=spFilter->filter(0);
			const SpatialVector root=
					spOutputVertices->spatialVector(rootIndex,0);

			for(I32 filterIndex=0;filterIndex<filterCount;filterIndex++)
			{
				const I32 reIndex=spFilter->filter(filterIndex);
				const U32 subCount=spInputVertices->subCount(reIndex);

				for(U32 subIndex=0;subIndex<subCount;subIndex++)
				{
					const I32 pointIndex=
							spOutputVertices->integer(reIndex,subIndex);
					if(pointSet.find(pointIndex)!=pointSet.end())
					{
						continue;
					}
					pointSet.insert(pointIndex);

					const SpatialVector point=
							spOutputVertices->spatialVector(reIndex,subIndex);
					spOutputVertices->set(reIndex,subIndex,
							root+0.2*(point-root));
				}
			}
		}
	}

#if FE_HGO_DEBUG
	feLog("HingeOp::handle done\n");
#endif

#if FE_HGO_PROFILE
	m_spProfileLoop->finish();
	m_spProfiler->end();

	feLog("%s:\n%s\n",m_spProfiler->name().c_str(),
			m_spProfiler->report().c_str());
#endif
}

void HingeOp::run(I32 a_id,sp<SpannedRange> a_spRange)
{
	const String collideMethod=catalog<String>("collideMethod");
	const I32 hitLimit=catalog<I32>("hitLimit");
	const Real threshold=catalog<Real>("threshold");
	const Real deadRadius=catalog<Real>("deadRadius");
	const Real liveRadius=catalog<Real>("liveRadius");
	const Real radianLimit=catalog<Real>("angleLimit")*degToRad;

	const BWORD useNearest=(collideMethod=="nearest");

	const SpatialVector axis=unitSafe(m_pivotB-m_pivotA);

	sp<SurfaceAccessorI> spOutputPoint;
	if(!access(spOutputPoint,m_spOutputAccessible,e_point,e_position)) return;

	Real& rMaxAngle=m_maxAngle[a_id<0? 0: a_id];

	for(SpannedRange::Iterator it=a_spRange->begin();!it.atEnd();it.step())
	{
		if(interrupted())
		{
			return;
		}

		const U32 pointIndex=it.value();

		const SpatialVector point=spOutputPoint->spatialVector(pointIndex);

#if FE_HGO_VERBOSE
		feLog("\npoint %d %s\n",pointIndex,c_print(point));
#endif

		const Real along=dot(point-m_pivotA,axis);
		const SpatialVector pivot=m_pivotA+axis*along;

		const SpatialVector from=point-pivot;
		const Real radius=magnitude(from);

		if(radius<=0.0 || radius<=deadRadius)
		{
			continue;
		}

		const SpatialVector fromUnit=from*(1.0/radius);
		const SpatialVector fromSide=cross(axis,fromUnit);

		const Real maxDistance= -1.0;
		const SpatialVector rayDir=unitSafe(fromSide);

#if FE_HGO_VERBOSE
		feLog("rayDir %s\n",c_print(rayDir));
#endif

		Array< sp<SurfaceI::ImpactI> > impactArray;
		U32 hitCount=0;
		U32 hitStart=0;

//		const SpatialVector rayOrigin=
//				useNearest? point: point-threshold*rayDir;
		const SpatialVector rayOrigin=point-threshold*rayDir;

		if(useNearest)
		{
			impactArray=m_spObstacle->nearestPoints(
					rayOrigin,maxDistance,hitLimit);
			hitCount=impactArray.size();
		}
		else
		{
			impactArray=m_spObstacle->rayImpacts(rayOrigin,rayDir,
					maxDistance,hitLimit);
			hitCount=impactArray.size();
			hitStart=hitCount-1;
		}

		for(U32 hitIndex=hitStart;hitIndex<hitCount;hitIndex++)
		{
			sp<SurfaceI::ImpactI> spImpact=impactArray[hitIndex];

			const SpatialVector intersection=spImpact->intersection();
			const I32 partitionIndex=spImpact->partitionIndex();

			const BWORD pushBack=(partitionIndex>=0 &&
					partitionIndex>m_partitionIndex);

			const SpatialVector reAxis=pushBack? -axis: axis;
			const SpatialVector reFromSide=pushBack? -fromSide: fromSide;
			const SpatialVector reRayDir=pushBack? -rayDir: rayDir;

#if FE_HGO_VERBOSE
			const SpatialVector hitNormal=spImpact->normal();
			feLog("intersect %s normal %s\n",
					c_print(intersection),c_print(hitNormal));
#endif

			Real alignRay=1.0;
			Real alignArm=1.0;
			SpatialVector goal;

			if(useNearest)
			{
				const SpatialVector direction=spImpact->direction();
				alignRay=dot(reRayDir,direction);
				if(alignRay<0.0)
				{
					continue;
				}

				const SpatialVector hitArm=unitSafe(intersection-pivot);
				alignArm=dot(fromUnit,hitArm);
				if(alignArm<0.0)
				{
					continue;
				}

				const Real distance=spImpact->distance();
				const SpatialVector projected=
						rayOrigin+distance*reRayDir*alignRay;

				goal=projected+threshold*reRayDir;
			}
			else
			{
				goal=intersection+threshold*reRayDir;
			}

			const SpatialVector toUnit=unitSafe(goal-pivot);

			//* TODO not always fully locked to axis
			const SpatialVector fromLocked=unitSafe(cross(reFromSide,reAxis));
			const SpatialVector toSide=cross(reAxis,toUnit);
			const SpatialVector toLocked=unitSafe(cross(toSide,reAxis));

#if FE_HGO_VERBOSE
			feLog("from %s to %s\n",
					c_print(fromLocked),c_print(toLocked));
#endif

			SpatialQuaternion rotation;
			set(rotation,fromLocked,toLocked);

			Real rotAngle=0.0;
			SpatialVector rotAxis;
			rotation.computeAngleAxis(rotAngle,rotAxis);

			const Real dotAxis=dot(reAxis,rotAxis);
			if(dotAxis<0.0)
			{
				rotAngle= -rotAngle;
			}

			Real fade=1.0;
			if(radius<liveRadius && liveRadius>deadRadius)
			{
				fade=(radius-deadRadius)/(liveRadius-deadRadius);
			}

			fade*=alignRay*alignArm;

#if FE_HGO_VERBOSE
			feLog("rotation %s\n",c_print(rotation));
			feLog("angle %.6G axis %s fade %.6G prev %.6G\n",
					rotAngle,c_print(rotAxis),fade,m_pointAngle[pointIndex]);
#endif

			rotAngle*=fade;

			const Real rotLimit=radianLimit-m_totalAngle[pointIndex];
			if(rotAngle>rotLimit)
			{
				rotAngle=rotLimit;
			}

			if(pushBack)
			{
				rotAngle= -rotAngle;

				if(m_pointAngle[pointIndex]>rotAngle)
				{
					m_pointAngle[pointIndex]=rotAngle;
				}
				else
				{
					rotAngle=m_pointAngle[pointIndex];
				}
			}
			else
			{
				if(rotAngle<0.0)
				{
					rotAngle=0.0;
				}

				if(m_pointAngle[pointIndex]<rotAngle)
				{
					m_pointAngle[pointIndex]=rotAngle;
				}
				else
				{
					rotAngle=m_pointAngle[pointIndex];
				}
			}

			if(rMaxAngle<rotAngle)
			{
				rMaxAngle=rotAngle;
			}

			if(m_spDrawGuide.isValid())
			{
				const Color red(1.0,0.0,0.0);
				const Color blue(0.0,0.0,1.0);
				const Color green(0.0,1.0,0.0);
				const Color yellow(1.0,1.0,0.0);
				const Color white(1.0,1.0,1.0);

				Color colors[2];
				colors[0]=blue;
				colors[1]=green;

				SpatialVector line[2];
				line[0]=point;
				line[1]=point+fromLocked;

				m_spDrawGuide->drawLines(line,NULL,2,
						DrawI::e_strip,TRUE,colors);

				colors[1]=yellow;
				line[1]=point+toLocked;

				m_spDrawGuide->drawLines(line,NULL,2,
						DrawI::e_strip,TRUE,colors);

				colors[0]=white;
				line[0]=goal;

				m_spDrawGuide->drawLines(line,NULL,2,
						DrawI::e_strip,TRUE,colors);

				colors[1]=red;
				line[1]=intersection;

				m_spDrawGuide->drawLines(line,NULL,2,
						DrawI::e_strip,TRUE,colors);
			}
		}

		if(!hitCount)
		{
			//* no hits

			if(rMaxAngle<m_pointAngle[pointIndex])
			{
				rMaxAngle=m_pointAngle[pointIndex];
			}
		}

		if(m_spDrawGuide.isValid())
		{
			const Color blue(0.0,0.0,1.0);
			const Color lightred(1.0,0.3,0.3);

			Color colors[2];
			colors[0]=blue;
			colors[1]=lightred;

			SpatialVector line[2];
			line[0]=point;
			line[1]=point+rayDir*radius*m_pointAngle[pointIndex];

			m_spDrawGuide->drawLines(line,NULL,2,
					DrawI::e_strip,TRUE,colors);
		}
	}
}

//* static
void HingeOp::calcConversion(SpatialTransform& a_rTransform,
	const Real a_angle,const SpatialVector& a_rAxis,
	const SpatialVector& a_rPivot)
{
	setIdentity(a_rTransform);

	translate(a_rTransform,a_rPivot);

	const SpatialQuaternion rotation(a_angle,a_rAxis);
	a_rTransform*=SpatialTransform(rotation);

	translate(a_rTransform,-a_rPivot);
}
