/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

#include "operator/operator.pmh"

namespace fe
{

void SurfaceCopyOp::initialize(void)
{
	catalog<String>("icon")="FE_alpha";

	catalog<bool>("rotate")=true;
	catalog<String>("rotate","label")="Rotate";
	catalog<bool>("rotate","joined")=true;

	catalog<bool>("scale")=true;
	catalog<String>("scale","label")="Scale";

	catalog<SpatialVector>("defaultTangent")=SpatialVector(1.0,0.0,0.0);
	catalog<String>("defaultTangent","label")="Default Tangent";
	catalog<String>("defaultTangent","enabler")="rotate";

	catalog<SpatialVector>("defaultNormal")=SpatialVector(0.0,1.0,0.0);
	catalog<String>("defaultNormal","label")="Default Normal";
	catalog<String>("defaultNormal","enabler")="rotate";

	catalog<SpatialVector>("defaultScale")=SpatialVector(1.0,1.0,1.0);
	catalog<String>("defaultScale","label")="Default Scale";
	catalog<String>("defaultScale","enabler")="scale";

	catalog<String>("tangentAttr")="tangent";
	catalog<String>("tangentAttr","label")="Tangent Attribute";
	catalog<String>("tangentAttr","enabler")="rotate";
	catalog<String>("tangentAttr","hint")="Local X vector.";

	catalog<String>("normalAttr")="N";
	catalog<String>("normalAttr","label")="Normal Attribute";
	catalog<String>("normalAttr","enabler")="rotate";
	catalog<String>("normalAttr","hint")="Local Y vector.";

	catalog<String>("scaleAttr")="scale";
	catalog<String>("scaleAttr","label")="Scale Attribute";
	catalog<String>("scaleAttr","enabler")="scale";
	catalog<String>("scaleAttr","hint")="Alters size.";

	catalog< sp<Component> >("Input Surface");

	catalog< sp<Component> >("Template Points");

	//* don't copy input to output
	catalog< sp<Component> >("Output Surface");
	catalog<String>("Output Surface","IO")="output";
}

void SurfaceCopyOp::handle(Record& a_rSignal)
{
	const BWORD doRotate=catalog<bool>("rotate");
	const BWORD doScale=catalog<bool>("scale");
	const SpatialVector defaultTangent=catalog<SpatialVector>("defaultTangent");
	const SpatialVector defaultNormal=catalog<SpatialVector>("defaultNormal");
	const SpatialVector defaultScale=catalog<SpatialVector>("defaultScale");
	const String tangentAttr=catalog<String>("tangentAttr");
	const String normalAttr=catalog<String>("normalAttr");
	const String scaleAttr=catalog<String>("scaleAttr");

	sp<SurfaceAccessibleI> spSurfaceOutput;
	if(!accessOutput(spSurfaceOutput,a_rSignal)) return;

	sp<SurfaceAccessibleI> spSurfaceInput;
	if(!access(spSurfaceInput,"Input Surface")) return;

	sp<SurfaceAccessorI> spTemplatePoint;
	if(!access(spTemplatePoint,"Template Points",e_point,e_position)) return;

	sp<SurfaceAccessorI> spTemplateTangent;
	sp<SurfaceAccessorI> spTemplateNormal;

	if(doRotate)
	{
		access(spTemplateTangent,"Template Points",
				e_point,tangentAttr,e_warning);
		access(spTemplateNormal,"Template Points",
				e_point,normalAttr,e_warning);
	}

	sp<SurfaceAccessorI> spTemplateScale;
	if(doScale)
	{
		access(spTemplateScale,"Template Points",
				e_point,scaleAttr,e_warning);
	}

	const U32 pointCount=spTemplatePoint->count();

	Array<SpatialTransform> transformArray;
	transformArray.resize(pointCount);

	SpatialVector norm=defaultNormal;
	SpatialVector tangent=defaultTangent;
	SpatialVector scaling=defaultScale;

	if(doRotate || doScale)
	{
		for(U32 pointIndex=0;pointIndex<pointCount;pointIndex++)
		{
			const SpatialVector point=
					spTemplatePoint->spatialVector(pointIndex);

			if(spTemplateTangent.isValid())
			{
				tangent=spTemplateTangent->spatialVector(pointIndex);
			}
			if(spTemplateNormal.isValid())
			{
				norm=spTemplateNormal->spatialVector(pointIndex);
			}
			if(spTemplateScale.isValid())
			{
				scaling=spTemplateScale->spatialVector(pointIndex);
			}

			SpatialTransform& rTransform=transformArray[pointIndex];

			makeFrameNormalY(rTransform,point,tangent,norm);

			if(doScale)
			{
				SpatialTransform scaleTransform;
				setIdentity(scaleTransform);
				scale(scaleTransform,scaling);

				rTransform=scaleTransform*rTransform;
			}
		}
	}
	else
	{
		for(U32 pointIndex=0;pointIndex<pointCount;pointIndex++)
		{
			const SpatialVector point=
					spTemplatePoint->spatialVector(pointIndex);

			SpatialTransform& rTransform=transformArray[pointIndex];

			setIdentity(rTransform);
			setTranslation(rTransform,point);
		}
	}

	spSurfaceOutput->instance(spSurfaceInput,transformArray);
}

} // namespace
